from __future__ import absolute_import, absolute_import, unicode_literals

import collections
import gzip
import json
import logging
import os
import re
import sys
import tempfile

import mopidy
from mopidy import compat, local, models
from mopidy.internal import encoding, timer
from mopidy.local import search, storage, translator

logger = logging.getLogger(__name__)


# TODO: move to load and dump in models?
def load_library(json_file):
    if not os.path.isfile(json_file):
        logger.info(
            'No local library metadata cache found at %s. Please run '
            '`mopidy local scan` to index your local music library. '
            'If you do not have a local music collection, you can disable the '
            'local backend to hide this message.',
            json_file)
        return {}
    try:
        with gzip.open(json_file, 'rb') as fp:
            return json.load(fp, object_hook=models.model_json_decoder)
    except (IOError, ValueError) as error:
        logger.warning(
            'Loading JSON local library failed: %s',
            encoding.locale_decode(error))
        return {}


def write_library(json_file, data):
    data['version'] = mopidy.__version__
    directory, basename = os.path.split(json_file)

    # TODO: cleanup directory/basename.* files.
    tmp = tempfile.NamedTemporaryFile(
        prefix=basename + '.', dir=directory, delete=False)

    try:
        with gzip.GzipFile(fileobj=tmp, mode='wb') as fp:
            json.dump(data, fp, cls=models.ModelJSONEncoder,
                      indent=2, separators=(',', ': '))
        os.rename(tmp.name, json_file)
    finally:
        if os.path.exists(tmp.name):
            os.remove(tmp.name)


class _BrowseCache(object):
    encoding = sys.getfilesystemencoding()
    splitpath_re = re.compile(r'([^/]+)')

    def __init__(self, uris):
        self._cache = {
            local.Library.ROOT_DIRECTORY_URI: collections.OrderedDict()}

        for track_uri in uris:
            path = translator.local_track_uri_to_path(track_uri, b'/')
            parts = self.splitpath_re.findall(
                path.decode(self.encoding, 'replace'))
            track_ref = models.Ref.track(uri=track_uri, name=parts.pop())

            # Look for our parents backwards as this is faster than having to
            # do a complete search for each add.
            parent_uri = None
            child = None
            for i in reversed(range(len(parts))):
                directory = '/'.join(parts[:i + 1])
                uri = translator.path_to_local_directory_uri(directory)

                # First dir we process is our parent
                if not parent_uri:
                    parent_uri = uri

                # We found ourselves and we exist, done.
                if uri in self._cache:
                    if child:
                        self._cache[uri][child.uri] = child
                    break

                # Initialize ourselves, store child if present, and add
                # ourselves as child for next loop.
                self._cache[uri] = collections.OrderedDict()
                if child:
                    self._cache[uri][child.uri] = child
                child = models.Ref.directory(uri=uri, name=parts[i])
            else:
                # Loop completed, so final child needs to be added to root.
                if child:
                    self._cache[
                        local.Library.ROOT_DIRECTORY_URI][child.uri] = child
                # If no parent was set we belong in the root.
                if not parent_uri:
                    parent_uri = local.Library.ROOT_DIRECTORY_URI

            self._cache[parent_uri][track_uri] = track_ref

    def lookup(self, uri):
        return self._cache.get(uri, {}).values()


class JsonLibrary(local.Library):
    name = 'json'

    def __init__(self, config):
        self._tracks = {}
        self._browse_cache = None
        self._media_dir = config['local']['media_dir']
        self._json_file = os.path.join(
            local.Extension.get_data_dir(config), b'library.json.gz')

        storage.check_dirs_and_files(config)

    def browse(self, uri):
        if not self._browse_cache:
            return []
        return self._browse_cache.lookup(uri)

    def load(self):
        logger.debug('Loading library: %s', self._json_file)
        with timer.time_logger('Loading tracks'):
            library = load_library(self._json_file)
            self._tracks = dict((t.uri, t) for t in library.get('tracks', []))
        with timer.time_logger('Building browse cache'):
            self._browse_cache = _BrowseCache(sorted(self._tracks.keys()))
        return len(self._tracks)

    def lookup(self, uri):
        try:
            return [self._tracks[uri]]
        except KeyError:
            return []

    def get_distinct(self, field, query=None):
        if field == 'track':
            def distinct(track):
                return {track.name}
        elif field == 'artist':
            def distinct(track):
                return {a.name for a in track.artists}
        elif field == 'albumartist':
            def distinct(track):
                album = track.album or models.Album()
                return {a.name for a in album.artists}
        elif field == 'album':
            def distinct(track):
                album = track.album or models.Album()
                return {album.name}
        elif field == 'composer':
            def distinct(track):
                return {a.name for a in track.composers}
        elif field == 'performer':
            def distinct(track):
                return {a.name for a in track.performers}
        elif field == 'date':
            def distinct(track):
                return {track.date}
        elif field == 'genre':
            def distinct(track):
                return {track.genre}
        else:
            return set()

        distinct_result = set()
        search_result = search.search(self._tracks.values(), query, limit=None)
        for track in search_result.tracks:
            distinct_result.update(distinct(track))
        return distinct_result - {None}

    def search(self, query=None, limit=100, offset=0, uris=None, exact=False):
        tracks = self._tracks.values()
        if exact:
            return search.find_exact(
                tracks, query=query, limit=limit, offset=offset, uris=uris)
        else:
            return search.search(
                tracks, query=query, limit=limit, offset=offset, uris=uris)

    def begin(self):
        return compat.itervalues(self._tracks)

    def add(self, track):
        self._tracks[track.uri] = track

    def remove(self, uri):
        self._tracks.pop(uri, None)

    def close(self):
        write_library(self._json_file, {'tracks': self._tracks.values()})

    def clear(self):
        try:
            os.remove(self._json_file)
            return True
        except OSError:
            return False



# encoding: utf-8

from __future__ import absolute_import, unicode_literals

import os
import platform
import shutil
import tempfile
import unittest

import pykka

from mopidy import core
from mopidy.internal import deprecation
from mopidy.m3u.backend import M3UBackend
from mopidy.models import Playlist, Track

from tests import dummy_audio, path_to_data_dir
from tests.m3u import generate_song


class M3UPlaylistsProviderTest(unittest.TestCase):
    backend_class = M3UBackend
    config = {
        'm3u': {
            'enabled': True,
            'base_dir': None,
            'default_encoding': 'latin-1',
            'default_extension': '.m3u',
            'playlists_dir': path_to_data_dir(''),
        }
    }

    def setUp(self):  # noqa: N802
        self.config['m3u']['playlists_dir'] = tempfile.mkdtemp()
        self.playlists_dir = self.config['m3u']['playlists_dir']
        self.base_dir = self.config['m3u']['base_dir'] or self.playlists_dir

        audio = dummy_audio.create_proxy()
        backend = M3UBackend.start(
            config=self.config, audio=audio).proxy()
        self.core = core.Core(backends=[backend])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

        if os.path.exists(self.playlists_dir):
            shutil.rmtree(self.playlists_dir)

    def test_created_playlist_is_persisted(self):
        uri = 'm3u:test.m3u'
        path = os.path.join(self.playlists_dir, b'test.m3u')
        self.assertFalse(os.path.exists(path))

        playlist = self.core.playlists.create('test')
        self.assertEqual('test', playlist.name)
        self.assertEqual(uri, playlist.uri)
        self.assertTrue(os.path.exists(path))

    def test_create_sanitizes_playlist_name(self):
        playlist = self.core.playlists.create('  ../../test FOO baR ')
        self.assertEqual('..|..|test FOO baR', playlist.name)
        path = os.path.join(self.playlists_dir, b'..|..|test FOO baR.m3u')
        self.assertEqual(self.playlists_dir, os.path.dirname(path))
        self.assertTrue(os.path.exists(path))

    def test_saved_playlist_is_persisted(self):
        uri1 = 'm3u:test1.m3u'
        uri2 = 'm3u:test2.m3u'

        path1 = os.path.join(self.playlists_dir, b'test1.m3u')
        path2 = os.path.join(self.playlists_dir, b'test2.m3u')

        playlist = self.core.playlists.create('test1')
        self.assertEqual('test1', playlist.name)
        self.assertEqual(uri1, playlist.uri)
        self.assertTrue(os.path.exists(path1))
        self.assertFalse(os.path.exists(path2))

        playlist = self.core.playlists.save(playlist.replace(name='test2'))
        self.assertEqual('test2', playlist.name)
        self.assertEqual(uri2, playlist.uri)
        self.assertFalse(os.path.exists(path1))
        self.assertTrue(os.path.exists(path2))

    def test_deleted_playlist_is_removed(self):
        uri = 'm3u:test.m3u'
        path = os.path.join(self.playlists_dir, b'test.m3u')

        self.assertFalse(os.path.exists(path))

        playlist = self.core.playlists.create('test')
        self.assertEqual('test', playlist.name)
        self.assertEqual(uri, playlist.uri)
        self.assertTrue(os.path.exists(path))

        self.core.playlists.delete(playlist.uri)
        self.assertFalse(os.path.exists(path))

    def test_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1))
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path) as f:
            contents = f.read()

        self.assertEqual(track.uri, contents.strip())

    def test_extended_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1), name='Test', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path) as f:
            m3u = f.read().splitlines()

        self.assertEqual(['#EXTM3U', '#EXTINF:-1,Test', track.uri], m3u)

    def test_latin1_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1), name='Test\x9f', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.copy(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path, 'rb') as f:
            m3u = f.read().splitlines()
        self.assertEqual([b'#EXTM3U', b'#EXTINF:-1,Test\x9f', track.uri], m3u)

    def test_utf8_playlist_contents_is_replaced_and_written_to_disk(self):
        track = Track(uri=generate_song(1), name='Test\u07b4', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.copy(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path, 'rb') as f:
            m3u = f.read().splitlines()
        self.assertEqual([b'#EXTM3U', b'#EXTINF:-1,Test?', track.uri], m3u)

    def test_playlists_are_loaded_at_startup(self):
        track = Track(uri='dummy:track:path2')
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup(playlist.uri)
        self.assertEqual(playlist.uri, result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual(track.uri, result.tracks[0].uri)

    def test_load_playlist_with_nonfilesystem_encoding_of_filename(self):
        path = os.path.join(self.playlists_dir, 'øæå.m3u'.encode('latin-1'))
        with open(path, 'wb+') as f:
            f.write(b'#EXTM3U\n')

        self.core.playlists.refresh()

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.as_list()
        if platform.system() == 'Darwin':
            self.assertEqual('%F8%E6%E5', result[0].name)
        else:
            self.assertEqual('\ufffd\ufffd\ufffd', result[0].name)

    @unittest.SkipTest
    def test_playlists_dir_is_created(self):
        pass

    def test_create_returns_playlist_with_name_set(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist.name, 'test')

    def test_create_returns_playlist_with_uri_set(self):
        playlist = self.core.playlists.create('test')
        self.assert_(playlist.uri)

    def test_create_adds_playlist_to_playlists_collection(self):
        playlist = self.core.playlists.create('test')
        playlists = self.core.playlists.as_list()
        self.assertIn(playlist.uri, [ref.uri for ref in playlists])

    def test_as_list_empty_to_start_with(self):
        self.assertEqual(len(self.core.playlists.as_list()), 0)

    def test_delete_non_existant_playlist(self):
        self.core.playlists.delete('m3u:unknown')

    def test_delete_playlist_removes_it_from_the_collection(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

        self.core.playlists.delete(playlist.uri)

        self.assertIsNone(self.core.playlists.lookup(playlist.uri))

    def test_delete_playlist_without_file(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

        path = os.path.join(self.playlists_dir, b'test.m3u')
        self.assertTrue(os.path.exists(path))

        os.remove(path)
        self.assertFalse(os.path.exists(path))

        self.core.playlists.delete(playlist.uri)
        self.assertIsNone(self.core.playlists.lookup(playlist.uri))

    def test_lookup_finds_playlist_by_uri(self):
        original_playlist = self.core.playlists.create('test')

        looked_up_playlist = self.core.playlists.lookup(original_playlist.uri)

        self.assertEqual(original_playlist, looked_up_playlist)

    def test_refresh(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

        self.core.playlists.refresh()

        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

    def test_save_replaces_existing_playlist_with_updated_playlist(self):
        playlist1 = self.core.playlists.create('test1')
        self.assertEqual(playlist1, self.core.playlists.lookup(playlist1.uri))

        playlist2 = playlist1.replace(name='test2')
        playlist2 = self.core.playlists.save(playlist2)
        self.assertIsNone(self.core.playlists.lookup(playlist1.uri))
        self.assertEqual(playlist2, self.core.playlists.lookup(playlist2.uri))

    def test_create_replaces_existing_playlist_with_updated_playlist(self):
        track = Track(uri=generate_song(1))
        playlist1 = self.core.playlists.create('test')
        playlist1 = self.core.playlists.save(playlist1.replace(tracks=[track]))
        self.assertEqual(playlist1, self.core.playlists.lookup(playlist1.uri))

        playlist2 = self.core.playlists.create('test')
        self.assertEqual(playlist1.uri, playlist2.uri)
        self.assertNotEqual(
            playlist1, self.core.playlists.lookup(playlist1.uri))
        self.assertEqual(playlist2, self.core.playlists.lookup(playlist1.uri))

    def test_save_playlist_with_new_uri(self):
        uri = 'm3u:test.m3u'
        self.core.playlists.save(Playlist(uri=uri))
        path = os.path.join(self.playlists_dir, b'test.m3u')
        self.assertTrue(os.path.exists(path))

    def test_playlist_with_unknown_track(self):
        track = Track(uri='file:///dev/null')
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup('m3u:test.m3u')
        self.assertEqual('m3u:test.m3u', result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual(track.uri, result.tracks[0].uri)

    def test_playlist_with_absolute_path(self):
        track = Track(uri='/tmp/test.mp3')
        filepath = b'/tmp/test.mp3'
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup('m3u:test.m3u')
        self.assertEqual('m3u:test.m3u', result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual('file://' + filepath, result.tracks[0].uri)

    def test_playlist_with_relative_path(self):
        track = Track(uri='test.mp3')
        filepath = os.path.join(self.base_dir, b'test.mp3')
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup('m3u:test.m3u')
        self.assertEqual('m3u:test.m3u', result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual('file://' + filepath, result.tracks[0].uri)

    def test_playlist_sort_order(self):
        def check_order(playlists, names):
            self.assertEqual(names, [playlist.name for playlist in playlists])

        self.core.playlists.create('c')
        self.core.playlists.create('a')
        self.core.playlists.create('b')

        check_order(self.core.playlists.as_list(), ['a', 'b', 'c'])

        self.core.playlists.refresh()

        check_order(self.core.playlists.as_list(), ['a', 'b', 'c'])

        playlist = self.core.playlists.lookup('m3u:a.m3u')
        playlist = playlist.replace(name='d')
        playlist = self.core.playlists.save(playlist)

        check_order(self.core.playlists.as_list(), ['b', 'c', 'd'])

        self.core.playlists.delete('m3u:c.m3u')

        check_order(self.core.playlists.as_list(), ['b', 'd'])

    def test_get_items_returns_item_refs(self):
        track = Track(uri='dummy:a', name='A', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))

        item_refs = self.core.playlists.get_items(playlist.uri)

        self.assertEqual(len(item_refs), 1)
        self.assertEqual(item_refs[0].type, 'track')
        self.assertEqual(item_refs[0].uri, 'dummy:a')
        self.assertEqual(item_refs[0].name, 'A')

    def test_get_items_of_unknown_playlist_returns_none(self):
        item_refs = self.core.playlists.get_items('dummy:unknown')

        self.assertIsNone(item_refs)


class M3UPlaylistsProviderBaseDirectoryTest(M3UPlaylistsProviderTest):

    def setUp(self):  # noqa: N802
        self.config['m3u']['base_dir'] = tempfile.mkdtemp()
        super(M3UPlaylistsProviderBaseDirectoryTest, self).setUp()


class DeprecatedM3UPlaylistsProviderTest(M3UPlaylistsProviderTest):

    def run(self, result=None):
        with deprecation.ignore(ids=['core.playlists.filter',
                                     'core.playlists.filter:kwargs_criteria',
                                     'core.playlists.get_playlists']):
            return super(DeprecatedM3UPlaylistsProviderTest, self).run(result)

    def test_filter_without_criteria(self):
        self.assertEqual(self.core.playlists.get_playlists(),
                         self.core.playlists.filter())

    def test_filter_with_wrong_criteria(self):
        self.assertEqual([], self.core.playlists.filter(name='foo'))

    def test_filter_with_right_criteria(self):
        playlist = self.core.playlists.create('test')
        playlists = self.core.playlists.filter(name='test')
        self.assertEqual([playlist], playlists)

    def test_filter_by_name_returns_single_match(self):
        self.core.playlists.create('a')
        playlist = self.core.playlists.create('b')

        self.assertEqual([playlist], self.core.playlists.filter(name='b'))

    def test_filter_by_name_returns_no_matches(self):
        self.core.playlists.create('a')
        self.core.playlists.create('b')

        self.assertEqual([], self.core.playlists.filter(name='c'))



from __future__ import absolute_import, unicode_literals

import random
import unittest

import pykka

from mopidy import core
from mopidy.core import PlaybackState
from mopidy.internal import deprecation
from mopidy.local import actor
from mopidy.models import Playlist, Track

from tests import dummy_audio, path_to_data_dir
from tests.local import generate_song, populate_tracklist


class LocalTracklistProviderTest(unittest.TestCase):
    config = {
        'core': {
            'data_dir': path_to_data_dir(''),
            'max_tracklist_length': 10000
        },
        'local': {
            'media_dir': path_to_data_dir(''),
            'playlists_dir': b'',
            'library': 'json',
        }
    }
    tracks = [
        Track(uri=generate_song(i), length=4464) for i in range(1, 4)]

    def run(self, result=None):
        with deprecation.ignore('core.tracklist.add:tracks_arg'):
            return super(LocalTracklistProviderTest, self).run(result)

    def setUp(self):  # noqa: N802
        self.audio = dummy_audio.create_proxy()
        self.backend = actor.LocalBackend.start(
            config=self.config, audio=self.audio).proxy()
        self.core = core.Core.start(audio=self.audio,
                                    backends=[self.backend],
                                    config=self.config).proxy()
        self.controller = self.core.tracklist
        self.playback = self.core.playback

        assert len(self.tracks) == 3, 'Need three tracks to run tests.'

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def assert_state_is(self, state):
        self.assertEqual(self.playback.get_state().get(), state)

    def assert_current_track_is(self, track):
        self.assertEqual(self.playback.get_current_track().get(), track)

    def test_length(self):
        self.assertEqual(0, len(self.controller.get_tl_tracks().get()))
        self.assertEqual(0, self.controller.get_length().get())
        self.controller.add(self.tracks)
        self.assertEqual(3, len(self.controller.get_tl_tracks().get()))
        self.assertEqual(3, self.controller.get_length().get())

    def test_add(self):
        for track in self.tracks:
            added = self.controller.add([track]).get()
            tracks = self.controller.get_tracks().get()
            tl_tracks = self.controller.get_tl_tracks().get()

            self.assertEqual(track, tracks[-1])
            self.assertEqual(added[0], tl_tracks[-1])
            self.assertEqual(track, added[0].track)

    def test_add_at_position(self):
        for track in self.tracks[:-1]:
            added = self.controller.add([track], 0).get()
            tracks = self.controller.get_tracks().get()
            tl_tracks = self.controller.get_tl_tracks().get()

            self.assertEqual(track, tracks[0])
            self.assertEqual(added[0], tl_tracks[0])
            self.assertEqual(track, added[0].track)

    @populate_tracklist
    def test_add_at_position_outside_of_playlist(self):
        for track in self.tracks:
            added = self.controller.add([track], len(self.tracks) + 2).get()
            tracks = self.controller.get_tracks().get()
            tl_tracks = self.controller.get_tl_tracks().get()

            self.assertEqual(track, tracks[-1])
            self.assertEqual(added[0], tl_tracks[-1])
            self.assertEqual(track, added[0].track)

    @populate_tracklist
    def test_filter_by_tlid(self):
        tl_track = self.controller.get_tl_tracks().get()[1]
        result = self.controller.filter({'tlid': [tl_track.tlid]}).get()
        self.assertEqual([tl_track], result)

    @populate_tracklist
    def test_filter_by_uri(self):
        tl_track = self.controller.get_tl_tracks().get()[1]
        result = self.controller.filter({'uri': [tl_track.track.uri]}).get()
        self.assertEqual([tl_track], result)

    @populate_tracklist
    def test_filter_by_uri_returns_nothing_for_invalid_uri(self):
        self.assertEqual([], self.controller.filter({'uri': ['foobar']}).get())

    def test_filter_by_uri_returns_single_match(self):
        t = Track(uri='a')
        self.controller.add([Track(uri='z'), t, Track(uri='y')])

        result = self.controller.filter({'uri': ['a']}).get()
        self.assertEqual(t, result[0].track)

    def test_filter_by_uri_returns_multiple_matches(self):
        track = Track(uri='a')
        self.controller.add([Track(uri='z'), track, track])
        tl_tracks = self.controller.filter({'uri': ['a']}).get()
        self.assertEqual(track, tl_tracks[0].track)
        self.assertEqual(track, tl_tracks[1].track)

    def test_filter_by_uri_returns_nothing_if_no_match(self):
        self.controller.playlist = Playlist(
            tracks=[Track(uri='z'), Track(uri='y')])
        self.assertEqual([], self.controller.filter({'uri': ['a']}).get())

    def test_filter_by_multiple_criteria_returns_elements_matching_all(self):
        t1 = Track(uri='a', name='x')
        t2 = Track(uri='b', name='x')
        t3 = Track(uri='b', name='y')
        self.controller.add([t1, t2, t3])

        result1 = self.controller.filter({'uri': ['a'], 'name': ['x']}).get()
        self.assertEqual(t1, result1[0].track)

        result2 = self.controller.filter({'uri': ['b'], 'name': ['x']}).get()
        self.assertEqual(t2, result2[0].track)

        result3 = self.controller.filter({'uri': ['b'], 'name': ['y']}).get()
        self.assertEqual(t3, result3[0].track)

    def test_filter_by_criteria_that_is_not_present_in_all_elements(self):
        track1 = Track()
        track2 = Track(uri='b')
        track3 = Track()

        self.controller.add([track1, track2, track3])
        result = self.controller.filter({'uri': ['b']}).get()
        self.assertEqual(track2, result[0].track)

    @populate_tracklist
    def test_clear(self):
        self.controller.clear().get()
        self.assertEqual(len(self.controller.get_tracks().get()), 0)

    def test_clear_empty_playlist(self):
        self.controller.clear().get()
        self.assertEqual(len(self.controller.get_tracks().get()), 0)

    @populate_tracklist
    def test_clear_when_playing(self):
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.controller.clear().get()
        self.assert_state_is(PlaybackState.STOPPED)

    def test_add_appends_to_the_tracklist(self):
        self.controller.add([Track(uri='a'), Track(uri='b')])

        tracks = self.controller.get_tracks().get()
        self.assertEqual(len(tracks), 2)

        self.controller.add([Track(uri='c'), Track(uri='d')])

        tracks = self.controller.get_tracks().get()
        self.assertEqual(len(tracks), 4)
        self.assertEqual(tracks[0].uri, 'a')
        self.assertEqual(tracks[1].uri, 'b')
        self.assertEqual(tracks[2].uri, 'c')
        self.assertEqual(tracks[3].uri, 'd')

    def test_add_does_not_reset_version(self):
        version = self.controller.get_version().get()
        self.controller.add([])
        self.assertEqual(self.controller.get_version().get(), version)

    @populate_tracklist
    def test_add_preserves_playing_state(self):
        self.playback.play().get()

        track = self.playback.get_current_track().get()
        tracks = self.controller.get_tracks().get()
        self.controller.add(tracks[1:2]).get()

        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_add_preserves_stopped_state(self):
        tracks = self.controller.get_tracks().get()
        self.controller.add(tracks[1:2]).get()

        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_add_returns_the_tl_tracks_that_was_added(self):
        tracks = self.controller.get_tracks().get()

        added = self.controller.add(tracks[1:2]).get()
        tracks = self.controller.get_tracks().get()
        self.assertEqual(added[0].track, tracks[1])

    @populate_tracklist
    def test_move_single(self):
        self.controller.move(0, 0, 2)

        tracks = self.controller.get_tracks().get()
        self.assertEqual(tracks[2], self.tracks[0])

    @populate_tracklist
    def test_move_group(self):
        self.controller.move(0, 2, 1)

        tracks = self.controller.get_tracks().get()
        self.assertEqual(tracks[1], self.tracks[0])
        self.assertEqual(tracks[2], self.tracks[1])

    @populate_tracklist
    def test_moving_track_outside_of_playlist(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.move(0, 0, num_tracks + 5).get()

    @populate_tracklist
    def test_move_group_outside_of_playlist(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.move(0, 2, num_tracks + 5).get()

    @populate_tracklist
    def test_move_group_out_of_range(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.move(num_tracks + 2, num_tracks + 3, 0).get()

    @populate_tracklist
    def test_move_group_invalid_group(self):
        with self.assertRaises(AssertionError):
            self.controller.move(2, 1, 0).get()

    def test_tracks_attribute_is_immutable(self):
        tracks1 = self.controller.tracks.get()
        tracks2 = self.controller.tracks.get()
        self.assertNotEqual(id(tracks1), id(tracks2))

    @populate_tracklist
    def test_remove(self):
        track1 = self.controller.get_tracks().get()[1]
        track2 = self.controller.get_tracks().get()[2]
        version = self.controller.get_version().get()
        self.controller.remove({'uri': [track1.uri]})
        self.assertLess(version, self.controller.get_version().get())
        self.assertNotIn(track1, self.controller.get_tracks().get())
        self.assertEqual(track2, self.controller.get_tracks().get()[1])

    @populate_tracklist
    def test_removing_track_that_does_not_exist_does_nothing(self):
        self.controller.remove({'uri': ['/nonexistant']}).get()

    def test_removing_from_empty_playlist_does_nothing(self):
        self.controller.remove({'uri': ['/nonexistant']}).get()

    @populate_tracklist
    def test_remove_lists(self):
        version = self.controller.get_version().get()
        tracks = self.controller.get_tracks().get()
        track0 = tracks[0]
        track1 = tracks[1]
        track2 = tracks[2]

        self.controller.remove({'uri': [track0.uri, track2.uri]})

        tracks = self.controller.get_tracks().get()
        self.assertLess(version, self.controller.get_version().get())
        self.assertNotIn(track0, tracks)
        self.assertNotIn(track2, tracks)
        self.assertEqual(track1, tracks[0])

    @populate_tracklist
    def test_shuffle(self):
        random.seed(1)
        self.controller.shuffle()

        shuffled_tracks = self.controller.get_tracks().get()

        self.assertNotEqual(self.tracks, shuffled_tracks)
        self.assertEqual(set(self.tracks), set(shuffled_tracks))

    @populate_tracklist
    def test_shuffle_subset(self):
        random.seed(1)
        self.controller.shuffle(1, 3)

        shuffled_tracks = self.controller.get_tracks().get()

        self.assertNotEqual(self.tracks, shuffled_tracks)
        self.assertEqual(self.tracks[0], shuffled_tracks[0])
        self.assertEqual(set(self.tracks), set(shuffled_tracks))

    @populate_tracklist
    def test_shuffle_invalid_subset(self):
        with self.assertRaises(AssertionError):
            self.controller.shuffle(3, 1).get()

    @populate_tracklist
    def test_shuffle_superset(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.shuffle(1, num_tracks + 5).get()

    @populate_tracklist
    def test_shuffle_open_subset(self):
        random.seed(1)
        self.controller.shuffle(1)

        shuffled_tracks = self.controller.get_tracks().get()

        self.assertNotEqual(self.tracks, shuffled_tracks)
        self.assertEqual(self.tracks[0], shuffled_tracks[0])
        self.assertEqual(set(self.tracks), set(shuffled_tracks))

    @populate_tracklist
    def test_slice_returns_a_subset_of_tracks(self):
        track_slice = self.controller.slice(1, 3).get()
        self.assertEqual(2, len(track_slice))
        self.assertEqual(self.tracks[1], track_slice[0].track)
        self.assertEqual(self.tracks[2], track_slice[1].track)

    @populate_tracklist
    def test_slice_returns_empty_list_if_indexes_outside_tracks_list(self):
        self.assertEqual(0, len(self.controller.slice(7, 8).get()))
        self.assertEqual(0, len(self.controller.slice(-1, 1).get()))

    def test_version_does_not_change_when_adding_nothing(self):
        version = self.controller.get_version().get()
        self.controller.add([])
        self.assertEqual(version, self.controller.get_version().get())

    def test_version_increases_when_adding_something(self):
        version = self.controller.get_version().get()
        self.controller.add([Track()])
        self.assertLess(version, self.controller.get_version().get())



from __future__ import absolute_import, unicode_literals

import time
import unittest

import mock

import pykka

from mopidy import core
from mopidy.core import PlaybackState
from mopidy.internal import deprecation
from mopidy.local import actor
from mopidy.models import TlTrack, Track

from tests import dummy_audio, path_to_data_dir
from tests.local import generate_song, populate_tracklist


# TODO Test 'playlist repeat', e.g. repeat=1,single=0


class LocalPlaybackProviderTest(unittest.TestCase):
    config = {
        'core': {
            'data_dir': path_to_data_dir(''),
            'max_tracklist_length': 10000,
        },
        'local': {
            'media_dir': path_to_data_dir(''),
            'library': 'json',
        }
    }

    # We need four tracks so that our shuffled track tests behave nicely with
    # reversed as a fake shuffle. Ensuring that shuffled order is [4,3,2,1] and
    # normal order [1,2,3,4] which means next_track != next_track_with_random
    tracks = [
        Track(uri=generate_song(i), length=4464) for i in (1, 2, 3, 4)]

    def add_track(self, uri):
        track = Track(uri=uri, length=4464)
        self.tracklist.add([track])

    def trigger_about_to_finish(self):
        # Flush any queued core calls.
        self.playback.get_current_tl_track().get()

        callback = self.audio.get_about_to_finish_callback().get()
        callback()

    def run(self, result=None):
        with deprecation.ignore('core.tracklist.add:tracks_arg'):
            return super(LocalPlaybackProviderTest, self).run(result)

    def setUp(self):  # noqa: N802
        self.audio = dummy_audio.create_proxy()
        self.backend = actor.LocalBackend.start(
            config=self.config, audio=self.audio).proxy()
        self.core = core.Core.start(audio=self.audio,
                                    backends=[self.backend],
                                    config=self.config).proxy()
        self.playback = self.core.playback
        self.tracklist = self.core.tracklist

        assert len(self.tracks) >= 3, \
            'Need at least three tracks to run tests.'
        assert self.tracks[0].length >= 2000, \
            'First song needs to be at least 2000 miliseconds'

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def assert_state_is(self, state):
        self.assertEqual(self.playback.get_state().get(), state)

    def assert_current_track_is(self, track):
        self.assertEqual(self.playback.get_current_track().get(), track)

    def assert_current_track_is_not(self, track):
        self.assertNotEqual(self.playback.get_current_track().get(), track)

    def assert_current_track_index_is(self, index):
        tl_track = self.playback.get_current_tl_track().get()
        self.assertEqual(self.tracklist.index(tl_track).get(), index)

    def assert_next_tl_track_is(self, tl_track):
        current = self.playback.get_current_tl_track().get()
        self.assertEqual(self.tracklist.next_track(current).get(), tl_track)

    def assert_next_tl_track_is_not(self, tl_track):
        current = self.playback.get_current_tl_track().get()
        self.assertNotEqual(self.tracklist.next_track(current).get(), tl_track)

    def assert_previous_tl_track_is(self, tl_track):
        current = self.playback.get_current_tl_track().get()
        previous = self.tracklist.previous_track(current).get()
        self.assertEqual(previous, tl_track)

    def assert_eot_tl_track_is(self, tl_track):
        current = self.playback.get_current_tl_track().get()
        self.assertEqual(self.tracklist.eot_track(current).get(), tl_track)

    def assert_eot_tl_track_is_not(self, tl_track):
        current = self.playback.get_current_tl_track().get()
        self.assertNotEqual(self.tracklist.eot_track(current).get(), tl_track)

    def test_uri_scheme(self):
        self.assertNotIn('file', self.core.uri_schemes.get())
        self.assertIn('local', self.core.uri_schemes.get())

    def test_play_mp3(self):
        self.add_track('local:track:blank.mp3')
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_play_ogg(self):
        self.add_track('local:track:blank.ogg')
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_play_flac(self):
        self.add_track('local:track:blank.flac')
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_play_uri_with_non_ascii_bytes(self):
        # Regression test: If trying to do .split(u':') on a bytestring, the
        # string will be decoded from ASCII to Unicode, which will crash on
        # non-ASCII strings, like the bytestring the following URI decodes to.
        self.add_track('local:track:12%20Doin%E2%80%99%20It%20Right.flac')
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_initial_state_is_stopped(self):
        self.assert_state_is(PlaybackState.STOPPED)

    def test_play_with_empty_playlist(self):
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.STOPPED)

    def test_play_with_empty_playlist_return_value(self):
        self.assertEqual(self.playback.play().get(), None)

    @populate_tracklist
    def test_play_state(self):
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_play_return_value(self):
        self.assertEqual(self.playback.play().get(), None)

    @populate_tracklist
    def test_play_track_state(self):
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_play_track_return_value(self):
        self.assertIsNone(self.playback.play(self.tl_tracks.get()[-1]).get())

    @populate_tracklist
    def test_play_when_playing(self):
        self.playback.play().get()
        track = self.playback.get_current_track().get()
        self.playback.play().get()
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_play_when_paused(self):
        self.playback.play().get()
        track = self.playback.get_current_track().get()
        self.playback.pause().get()
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_play_when_paused_after_next(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.next().get()
        track = self.playback.get_current_track().get()
        self.playback.pause().get()
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_play_sets_current_track(self):
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_play_track_sets_current_track(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.assert_current_track_is(self.tracks[-1])

    @populate_tracklist
    def test_play_skips_to_next_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        uri = self.backend.playback.translate_uri(self.tracks[0].uri).get()
        self.audio.trigger_fake_playback_failure(uri)

        self.playback.play().get()
        self.assert_current_track_is_not(self.tracks[0])
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_current_track_after_completed_playlist(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.playback.next().get()
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_previous(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.previous().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_previous_more(self):
        self.playback.play().get()  # At track 0
        self.playback.next().get()  # At track 1
        self.playback.next().get()  # At track 2
        self.playback.previous().get()  # At track 1
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_previous_return_value(self):
        self.playback.play().get()
        self.playback.next().get()
        self.assertIsNone(self.playback.previous().get())

    @populate_tracklist
    def test_previous_does_not_trigger_playback(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.stop()
        self.playback.previous().get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_previous_at_start_of_playlist(self):
        self.playback.previous().get()
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    def test_previous_for_empty_playlist(self):
        self.playback.previous().get()
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_previous_skips_to_previous_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        uri = self.backend.playback.translate_uri(self.tracks[1].uri).get()
        self.audio.trigger_fake_playback_failure(uri)

        self.playback.play(self.tl_tracks.get()[2]).get()
        self.assert_current_track_is(self.tracks[2])
        self.playback.previous().get()
        self.assert_current_track_is_not(self.tracks[1])
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_next(self):
        self.playback.play().get()

        old_track = self.playback.get_current_track().get()
        old_position = self.tracklist.index().get()

        self.playback.next().get()

        self.assertEqual(self.tracklist.index().get(), old_position + 1)
        self.assert_current_track_is_not(old_track)

    @populate_tracklist
    def test_next_return_value(self):
        self.playback.play().get()
        self.assertEqual(self.playback.next().get(), None)

    @populate_tracklist
    def test_next_does_not_trigger_playback(self):
        self.playback.next().get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_next_at_end_of_playlist(self):
        self.playback.play().get()

        for i, track in enumerate(self.tracks):
            self.assert_state_is(PlaybackState.PLAYING)
            self.assert_current_track_is(track)
            self.assertEqual(self.tracklist.index().get(), i)

            self.playback.next()

        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_next_until_end_of_playlist_and_play_from_start(self):
        self.playback.play().get()

        for _ in self.tracks:
            self.playback.next().get()

        self.assert_current_track_is(None)
        self.assert_state_is(PlaybackState.STOPPED)

        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(self.tracks[0])

    def test_next_for_empty_playlist(self):
        self.playback.next().get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_next_skips_to_next_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        uri = self.backend.playback.translate_uri(self.tracks[1].uri).get()
        self.audio.trigger_fake_playback_failure(uri)

        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.playback.next().get()
        self.assert_current_track_is_not(self.tracks[1])
        self.assert_current_track_is(self.tracks[2])

    @populate_tracklist
    def test_next_track_before_play(self):
        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    def test_next_track_during_play(self):
        self.playback.play().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    @populate_tracklist
    def test_next_track_after_previous(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.previous().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    def test_next_track_empty_playlist(self):
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_next_track_at_end_of_playlist(self):
        self.playback.play().get()
        for _ in self.tl_tracks.get()[1:]:
            self.playback.next().get()
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_next_track_at_end_of_playlist_with_repeat(self):
        self.tracklist.repeat = True
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.playback.next().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_next_track_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        self.assert_next_tl_track_is(self.tl_tracks.get()[-1])

    @populate_tracklist
    def test_next_with_consume(self):
        self.tracklist.consume = True
        self.playback.play().get()
        self.playback.next().get()
        self.assertNotIn(self.tracks[0], self.tracklist.get_tracks().get())

    @populate_tracklist
    def test_next_with_single_and_repeat(self):
        self.tracklist.single = True
        self.tracklist.repeat = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.playback.next().get()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_next_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[-1])
        self.playback.next().get()
        self.assert_current_track_is(self.tracks[-2])

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_next_track_with_random_after_append_playlist(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        current_tl_track = self.playback.get_current_tl_track().get()

        expected_tl_track = self.tl_tracks.get()[-1]
        next_tl_track = self.tracklist.next_track(current_tl_track).get()

        # Baseline checking that first next_track is last tl track per our fake
        # shuffle.
        self.assertEqual(next_tl_track, expected_tl_track)

        self.tracklist.add(self.tracks[:1])

        old_next_tl_track = next_tl_track
        expected_tl_track = self.tracklist.tl_tracks.get()[-1]
        next_tl_track = self.tracklist.next_track(current_tl_track).get()

        # Verify that first next track has changed since we added to the
        # playlist.
        self.assertEqual(next_tl_track, expected_tl_track)
        self.assertNotEqual(next_tl_track, old_next_tl_track)

    @populate_tracklist
    def test_end_of_track(self):
        self.playback.play().get()

        old_track = self.playback.get_current_track().get()
        old_position = self.tracklist.index().get()

        self.trigger_about_to_finish()

        new_track = self.playback.get_current_track().get()
        self.assertEqual(self.tracklist.index().get(), old_position + 1)
        self.assertNotEqual(new_track.uri, old_track.uri)

    @populate_tracklist
    def test_end_of_track_return_value(self):
        self.playback.play().get()
        self.assertEqual(self.trigger_about_to_finish(), None)

    @populate_tracklist
    def test_end_of_track_does_not_trigger_playback(self):
        self.trigger_about_to_finish()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_track_at_end_of_playlist(self):
        self.playback.play().get()

        for i, track in enumerate(self.tracks):
            self.assert_state_is(PlaybackState.PLAYING)
            self.assert_current_track_is(track)
            self.assertEqual(self.tracklist.index().get(), i)

            self.trigger_about_to_finish()

        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_track_until_end_of_playlist_and_play_from_start(self):
        self.playback.play().get()

        for _ in self.tracks:
            self.trigger_about_to_finish()

        self.assertEqual(self.playback.get_current_track().get(), None)
        self.assert_state_is(PlaybackState.STOPPED)

        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(self.tracks[0])

    def test_end_of_track_for_empty_playlist(self):
        self.trigger_about_to_finish()
        self.assert_state_is(PlaybackState.STOPPED)

    # TODO: On about to finish does not handle skipping to next track yet.
    @unittest.expectedFailure
    @populate_tracklist
    def test_end_of_track_skips_to_next_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        return_values = [True, False, True]
        self.backend.playback.play = lambda: return_values.pop()
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.trigger_about_to_finish()
        self.assert_current_track_is_not(self.tracks[1])
        self.assert_current_track_is(self.tracks[2])

    @populate_tracklist
    def test_end_of_track_track_before_play(self):
        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    def test_end_of_track_track_during_play(self):
        self.playback.play().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    @populate_tracklist
    def test_about_to_finish_after_previous(self):
        self.playback.play().get()
        self.trigger_about_to_finish()
        self.playback.previous().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    def test_end_of_track_track_empty_playlist(self):
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_end_of_track_track_at_end_of_playlist(self):
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.trigger_about_to_finish()

        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_end_of_track_track_at_end_of_playlist_with_repeat(self):
        self.tracklist.repeat = True
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.trigger_about_to_finish()

        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_end_of_track_track_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        self.assert_next_tl_track_is(self.tl_tracks.get()[-1])

    @populate_tracklist
    def test_end_of_track_with_consume(self):
        self.tracklist.consume = True
        self.playback.play().get()
        self.trigger_about_to_finish()
        self.assertNotIn(self.tracks[0], self.tracklist.get_tracks().get())

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_end_of_track_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[-1])
        self.trigger_about_to_finish()
        self.assert_current_track_is(self.tracks[-2])

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_end_of_track_track_with_random_after_append_playlist(
            self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        current_tl_track = self.playback.get_current_tl_track().get()

        expected_tl_track = self.tracklist.get_tl_tracks().get()[-1]
        eot_tl_track = self.tracklist.eot_track(current_tl_track).get()

        # Baseline checking that first eot_track is last tl track per our fake
        # shuffle.
        self.assertEqual(eot_tl_track, expected_tl_track)

        self.tracklist.add(self.tracks[:1])

        old_eot_tl_track = eot_tl_track
        expected_tl_track = self.tracklist.get_tl_tracks().get()[-1]
        eot_tl_track = self.tracklist.eot_track(current_tl_track).get()

        # Verify that first next track has changed since we added to the
        # playlist.
        self.assertEqual(eot_tl_track, expected_tl_track)
        self.assertNotEqual(eot_tl_track, old_eot_tl_track)

    @populate_tracklist
    def test_previous_track_before_play(self):
        self.assert_previous_tl_track_is(None)

    @populate_tracklist
    def test_previous_track_after_play(self):
        self.playback.play().get()
        self.assert_previous_tl_track_is(None)

    @populate_tracklist
    def test_previous_track_after_next(self):
        self.playback.play().get()
        self.playback.next().get()
        self.assert_previous_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    def test_previous_track_after_previous(self):
        self.playback.play().get()  # At track 0
        self.playback.next().get()  # At track 1
        self.playback.next().get()  # At track 2
        self.playback.previous().get()  # At track 1
        self.assert_previous_tl_track_is(self.tl_tracks.get()[0])

    def test_previous_track_empty_playlist(self):
        self.assert_previous_tl_track_is(None)

    @populate_tracklist
    def test_previous_track_with_consume(self):
        self.tracklist.consume = True
        for _ in self.tracks:
            self.playback.next()
            current = self.playback.get_current_tl_track().get()
            self.assert_previous_tl_track_is(current)

    @populate_tracklist
    def test_previous_track_with_random(self):
        self.tracklist.random = True
        for _ in self.tracks:
            self.playback.next()
            current = self.playback.get_current_tl_track().get()
            self.assert_previous_tl_track_is(current)

    @populate_tracklist
    def test_initial_current_track(self):
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_current_track_during_play(self):
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_current_track_after_next(self):
        self.playback.play()
        self.playback.next().get()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_initial_tracklist_position(self):
        self.assertEqual(self.tracklist.index().get(), None)

    @populate_tracklist
    def test_tracklist_position_during_play(self):
        self.playback.play().get()
        self.assert_current_track_index_is(0)

    @populate_tracklist
    def test_tracklist_position_after_next(self):
        self.playback.play().get()
        self.playback.next().get()
        self.assert_current_track_index_is(1)

    @populate_tracklist
    def test_tracklist_position_at_end_of_playlist(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_current_track_index_is(None)

    @mock.patch('mopidy.core.playback.PlaybackController._on_tracklist_change')
    def test_on_tracklist_change_gets_called(self, change_mock):
        self.tracklist.add([Track()]).get()
        change_mock.assert_called_once_with()

    @populate_tracklist
    def test_on_tracklist_change_when_playing(self):
        self.playback.play().get()
        current_track = self.playback.get_current_track().get()
        self.tracklist.add([self.tracks[2]])
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_on_tracklist_change_when_stopped(self):
        self.tracklist.add([self.tracks[2]])
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_on_tracklist_change_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        current_track = self.playback.get_current_track().get()
        self.tracklist.add([self.tracks[2]])
        self.assert_state_is(PlaybackState.PAUSED)
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_pause_when_stopped(self):
        self.playback.pause()
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_pause_when_playing(self):
        self.playback.play().get()
        self.playback.pause()
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_pause_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        self.playback.pause()
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_pause_return_value(self):
        self.playback.play().get()
        self.assertIsNone(self.playback.pause().get())

    @populate_tracklist
    def test_resume_when_stopped(self):
        self.playback.resume()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_resume_when_playing(self):
        self.playback.play().get()
        self.playback.resume()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_resume_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        self.playback.resume()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_resume_return_value(self):
        self.playback.play().get()
        self.playback.pause()
        self.assertIsNone(self.playback.resume().get())

    @unittest.SkipTest  # Uses sleep and might not work with LocalBackend
    @populate_tracklist
    def test_resume_continues_from_right_position(self):
        self.playback.play().get()
        time.sleep(0.2)
        self.playback.pause()
        self.playback.resume()
        self.assertNotEqual(self.playback.time_position, 0)

    @populate_tracklist
    def test_seek_when_stopped(self):
        result = self.playback.seek(1000)
        self.assert_(result, 'Seek return value was %s' % result)

    @populate_tracklist
    def test_seek_when_stopped_updates_position(self):
        self.playback.seek(1000).get()
        position = self.playback.time_position
        self.assertGreaterEqual(position, 990)

    def test_seek_on_empty_playlist(self):
        self.assertFalse(self.playback.seek(0).get())

    def test_seek_on_empty_playlist_updates_position(self):
        self.playback.seek(0).get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_seek_when_stopped_triggers_play(self):
        self.playback.seek(0).get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_seek_when_playing(self):
        self.playback.play().get()
        result = self.playback.seek(self.tracks[0].length - 1000)
        self.assert_(result, 'Seek return value was %s' % result)

    @populate_tracklist
    def test_seek_when_playing_updates_position(self):
        length = self.tracks[0].length
        self.playback.play().get()
        self.playback.seek(length - 1000).get()
        position = self.playback.get_time_position().get()
        self.assertGreaterEqual(position, length - 1010)

    @populate_tracklist
    def test_seek_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        result = self.playback.seek(self.tracks[0].length - 1000)
        self.assert_(result, 'Seek return value was %s' % result)
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_seek_when_paused_updates_position(self):
        length = self.tracks[0].length
        self.playback.play().get()
        self.playback.pause()
        self.playback.seek(length - 1000)
        position = self.playback.get_time_position().get()
        self.assertGreaterEqual(position, length - 1010)

    @unittest.SkipTest
    @populate_tracklist
    def test_seek_beyond_end_of_song(self):
        # FIXME need to decide return value
        self.playback.play().get()
        result = self.playback.seek(self.tracks[0].length * 100)
        self.assert_(not result, 'Seek return value was %s' % result)

    @populate_tracklist
    def test_seek_beyond_end_of_song_jumps_to_next_song(self):
        self.playback.play().get()
        self.playback.seek(self.tracks[0].length * 100).get()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_seek_beyond_end_of_song_for_last_track(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.playback.seek(self.tracks[-1].length * 100)
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_stop_when_stopped(self):
        self.playback.stop()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_stop_when_playing(self):
        self.playback.play().get()
        self.playback.stop()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_stop_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        self.playback.stop()
        self.assert_state_is(PlaybackState.STOPPED)

    def test_stop_return_value(self):
        self.playback.play().get()
        self.assertIsNone(self.playback.stop().get())

    def test_time_position_when_stopped(self):
        self.assertEqual(self.playback.get_time_position().get(), 0)

    @populate_tracklist
    def test_time_position_when_stopped_with_playlist(self):
        self.assertEqual(self.playback.get_time_position().get(), 0)

    @unittest.SkipTest  # Uses sleep and does might not work with LocalBackend
    @populate_tracklist
    def test_time_position_when_playing(self):
        self.playback.play().get()
        first = self.playback.time_position
        time.sleep(1)
        second = self.playback.time_position
        self.assertGreater(second, first)

    @populate_tracklist
    def test_time_position_when_paused(self):
        self.playback.play().get()
        self.playback.pause().get()
        first = self.playback.get_time_position().get()
        second = self.playback.get_time_position().get()
        self.assertEqual(first, second)

    @populate_tracklist
    def test_play_with_consume(self):
        self.tracklist.consume = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_playlist_is_empty_after_all_tracks_are_played_with_consume(self):
        self.tracklist.consume = True
        self.playback.play().get()

        for t in self.tracks:
            self.trigger_about_to_finish()
        # EOS should have trigger

        self.assertEqual(len(self.tracklist.get_tracks().get()), 0)

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_play_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[-1])

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_previous_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.random = True
        self.playback.play().get()
        self.playback.next().get()
        current_track = self.playback.get_current_track().get()
        self.playback.previous()
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_end_of_song_starts_next_track(self):
        self.playback.play().get()
        self.trigger_about_to_finish()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_end_of_song_with_single_and_repeat_starts_same(self):
        self.tracklist.single = True
        self.tracklist.repeat = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.trigger_about_to_finish()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_end_of_song_with_single_random_and_repeat_starts_same(self):
        self.tracklist.single = True
        self.tracklist.repeat = True
        self.tracklist.random = True
        self.playback.play().get()
        current_track = self.playback.get_current_track().get()
        self.trigger_about_to_finish()
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_end_of_song_with_single_stops(self):
        self.tracklist.single = True
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.trigger_about_to_finish()
        self.assert_current_track_is(None)
        # EOS should have triggered
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_song_with_single_and_random_stops(self):
        self.tracklist.single = True
        self.tracklist.random = True
        self.playback.play().get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_current_track_is(None)
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_playlist_stops(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_state_is(PlaybackState.STOPPED)

    def test_repeat_off_by_default(self):
        self.assertEqual(self.tracklist.get_repeat().get(), False)

    def test_random_off_by_default(self):
        self.assertEqual(self.tracklist.get_random().get(), False)

    def test_consume_off_by_default(self):
        self.assertEqual(self.tracklist.get_consume().get(), False)

    @populate_tracklist
    def test_random_until_end_of_playlist(self):
        self.tracklist.random = True
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.playback.next().get()
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_random_with_eot_until_end_of_playlist(self):
        self.tracklist.random = True
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.trigger_about_to_finish()

        self.assert_eot_tl_track_is(None)

    @populate_tracklist
    def test_random_until_end_of_playlist_and_play_from_start(self):
        self.tracklist.random = True
        self.playback.play().get()
        for _ in self.tracks:
            self.playback.next().get()
        self.assert_next_tl_track_is_not(None)
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_random_with_eot_until_end_of_playlist_and_play_from_start(self):
        self.tracklist.random = True
        self.playback.play().get()
        for _ in self.tracks:
            self.trigger_about_to_finish()
        # EOS should have triggered

        self.assert_eot_tl_track_is_not(None)
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_random_until_end_of_playlist_with_repeat(self):
        self.tracklist.repeat = True
        self.tracklist.random = True
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.playback.next()
        self.assert_next_tl_track_is_not(None)

    @populate_tracklist
    def test_played_track_during_random_not_played_again(self):
        self.tracklist.random = True
        self.playback.play().get()
        played = []
        for _ in self.tracks:
            track = self.playback.get_current_track().get()
            self.assertNotIn(track, played)
            played.append(track)
            self.playback.next().get()

    @populate_tracklist
    @mock.patch('random.shuffle')
    def test_play_track_then_enable_random(self, shuffle_mock):
        # Covers underlying issue IssueGH17RegressionTest tests for.
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        expected = self.tl_tracks.get()[::-1] + [None]
        actual = []

        self.playback.play().get()
        self.tracklist.random = True
        while self.playback.get_state().get() != PlaybackState.STOPPED:
            self.playback.next().get()
            actual.append(self.playback.get_current_tl_track().get())
            if len(actual) > len(expected):
                break
        self.assertEqual(actual, expected)

    @populate_tracklist
    def test_playing_track_that_isnt_in_playlist(self):
        with self.assertRaises(AssertionError):
            self.playback.play(TlTrack(17, Track())).get()
